// Betriebssysteme1 - Matthias Jauernig, 2005
// Empfnger-Programm der 13. bung
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <ulimit.h>

bool ende = false;

void sigintfkt(int signr){
	printf("Bereite Empfaenger-Ende vor...\n");
	ende = true;
}

int main(void){
	int pipesOffen = 0;
	int rc, i, fd, maxfd = 0;
	fd_set pipefd, pipefdcp;
	FILE *pipe[5];
	char eab[100], pipename[10];
	bool exists[5], pipeOffen[5];
	struct stat buf;
	struct timeval timeout;
	timerclear(&timeout);
	
	// -1.) Initialwerte fr Arrays festlegen
	for(i=0; i<5; i++){
		exists[i] = false;
		pipeOffen[i] = false;
		pipe[i] = NULL;
	}
	FD_ZERO(&pipefd);
	
	// 0.) Signalfunktion fr SIGINT (Ctrl-C) bestimmen
	signal(SIGINT, sigintfkt);
	
	printf("Warte auf Sender...\n");
	// 1.) prfe stndig die relevanten Pipenamen ab und lese von ihnen
	while(!ende || pipesOffen>0){
		// 1.1) auf noch nicht geffnete Pipes prfen und diese ffnen
		for(i=0; i<5; i++){
			// 1.1.1) generiere Pipenamen
			sprintf(pipename, "p%d", i);
			// 1.1.2) prfe, ob Pipe existiert
			rc = stat(pipename, &buf);
			if(rc < 0){ // Datei existiert nicht oder Fehler
				if(errno != ENOENT){ // Fehler bei stat
					perror("stat");
				}
			}else{ // Datei existiert bereits
				if(!S_ISFIFO(buf.st_mode) && !exists[i]){ // Datei ist kein FIFO
					printf("File %s existiert, ist aber keine Pipe!\n", pipename);
					exists[i] = true;
				}else{
					if(!pipeOffen[i]){ // Datei ist FIFO und noch nicht geffnet
						pipe[i] = fopen(pipename, "r");
						if(pipe[i] == NULL){
							perror("fopen");
						}else{
							// bei ende Pipe beseitigen
							if(ende){
								if(fclose(pipe[i]) < 0){
									perror("fclose");
								}
								rc = unlink(pipename);
								if(rc < 0){
									perror("unlink");
								}
							}else{
								pipesOffen++;
								pipeOffen[i] = true;
								printf("Sender '%s' hat sich angemeldet (%d"
								  " insgesamt)...\n", pipename, pipesOffen);
								// Filedeskriptor in fd_set eintragen
								fd = fileno(pipe[i]);
								FD_SET(fd, &pipefd);
							}
						}
					}
				}
			}
		} //end for

		// 1.2) mit select prfen, von welchen FDs gelesen werden kann
		pipefdcp = pipefd;
		maxfd = ulimit(4);
		rc = select(maxfd, &pipefdcp, 0, 0, &timeout);
		if(rc == 0){ // Timeout beendet
			continue;
		}
		if(rc < 0){ // Fehler
			perror("select");
			exit(1);
		}
		
		// 1.3) von den FDs lesen, an denen Daten anliegen
		for(i=0; i<5; i++){
			if(pipeOffen[i]){ // Pipe geffnet
				fd = fileno(pipe[i]);
				if(FD_ISSET(fd, &pipefdcp)){ // Daten liegen an
					rc = read(fd, eab, sizeof(eab));
					if(rc==0){ // Pipe geschlossen
					        pipeOffen[i] = false;
						// in Original-fd_set fd-Bit lschen
						FD_CLR(fd, &pipefd);
						sprintf(pipename, "p%d", i);
						printf("Sender '%s' hat sich abgemeldet "
						  "(%d brig)...\n", pipename, pipesOffen-1);
						if(fclose(pipe[i]) < 0){
							perror("fclose");
						}else{ // Pipe beseitigen
							pipesOffen--;
							exists[i] = false;
							rc = unlink(pipename);
							if(rc < 0){
								perror("unlink");
							}
						}
					}else if(rc > 0){ // Daten empfangen, also ausgeben
						printf("Pipe %d: '%s'\n",i,eab);
					}
				}
			}
		}
	}
	printf("Keine Pipe mehr offen, mache Schluss...\n");
	return 0;
}
